namespace Yosys { extern const char *yosys_version_str; const char *yosys_version_str="Yosys 0.65 (git sha1 b85cad634-dirty, g++ 16.1.1 -march=x86-64 -mtune=generic -O2 -fno-plt -fexceptions -fstack-clash-protection -fcf-protection -fno-omit-frame-pointer -mno-omit-leaf-frame-pointer -ffile-prefix-map=/build/yosys/src=/usr/src/debug/yosys -fPIC -O3) [startdir/yosys at makepkg]"; const char *yosys_git_hash_str="b85cad634-dirty"; }
